@@ -1,7 +1,7 @@
 /*
  * drawboard.c
  *
- * by Gary Wong, 1999-2001
+ * by Gary Wong <gtw@gnu.org>, 1999, 2000, 2001, 2002
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of version 2 of the GNU General Public License as
@@ -16,7 +16,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
- * $Id: drawboard.c,v 1.14 2001/10/29 15:13:18 gtw Exp $
+ * $Id: drawboard.c,v 1.32 2003/10/14 16:43:33 jsegrave Exp $
  */
 
 #include "config.h"
@@ -30,14 +30,16 @@
 
 #include "drawboard.h"
 #include "positionid.h"
+#include "i18n.h"
 
 int fClockwise = FALSE; /* Player 1 moves clockwise */
 
 /*
- *  GNU Backgammon  Position ID: 0123456789ABCD
- *  +13-14-15-16-17-18------19-20-21-22-23-24-+     O: gnubg (0 points)
- *  |                  |   | O  O  O  O     O | OO  Cube: 2
- *  |                  |   | O     O          | OO  On roll
+ *  GNU Backgammon  Position ID: 0123456789ABCD 
+ *                  Match ID   : 0123456789ABCD
+ *  +13-14-15-16-17-18------19-20-21-22-23-24-+     O: gnubg (Cube: 2)
+ *  |                  |   | O  O  O  O     O | OO  0 points
+ *  |                  |   | O     O          | OO  Cube offered at 2
  *  |                  |   |       O          | O
  *  |                  |   |                  | O
  *  |                  |   |                  | O   
@@ -46,16 +48,17 @@ int fClockwise = FALSE; /* Player 1 moves clockwise */
  *  |                  |   |                  | X
  *  |                  |   |                  | X
  *  |                  |   |       X  X  X  X | X   Rolled 11
- *  |                  |   |    X  X  X  X  X | XX  Cube: 2
- *  +12-11-10--9--8--7-------6--5--4--3--2--1-+     X: Gary (0 points)
+ *  |                  |   |    X  X  X  X  X | XX  0 points
+ *  +12-11-10--9--8--7-------6--5--4--3--2--1-+     X: Gary (Cube: 2)
  *
  */
 
 static char *DrawBoardStd( char *sz, int anBoard[ 2 ][ 25 ], int fRoll,
-                           char *asz[] ) {
+                           char *asz[], char *szMatchID, 
+                           int nChequers ) {
 
     char *pch = sz, *pchIn;
-    int x, y, an[ 2 ][ 25 ], cOffO = 15, cOffX = 15;
+    int x, y, an[ 2 ][ 25 ], cOffO = nChequers, cOffX = nChequers;
     static char achX[ 16 ] = "     X6789ABCDEF",
         achO[ 16 ] = "     O6789ABCDEF";
 
@@ -64,7 +67,7 @@ static char *DrawBoardStd( char *sz, int anBoard[ 2 ][ 25 ], int fRoll,
         cOffX -= anBoard[ 1 ][ x ];
     }
     
-    strcpy( pch, " GNU Backgammon  Position ID: " );
+    strcpy( pch, _ (" GNU Backgammon  Position ID: ") );
 
     pch += 30;
 
@@ -81,6 +84,13 @@ static char *DrawBoardStd( char *sz, int anBoard[ 2 ][ 25 ], int fRoll,
     
     pch += 14;
     *pch++ = '\n';
+
+    /* match id */
+
+    if ( szMatchID && *szMatchID ) {
+      sprintf ( pch, _("                 Match ID   : %s\n"), szMatchID );
+      pch = strchr ( pch, 0 );
+    }
             
     strcpy( pch, fRoll ? " +13-14-15-16-17-18------19-20-21-22-23-24-+     " :
             " +12-11-10--9--8--7-------6--5--4--3--2--1-+     " );
@@ -127,6 +137,11 @@ static char *DrawBoardStd( char *sz, int anBoard[ 2 ][ 25 ], int fRoll,
         if( y < 2 && asz[ y + 1 ] )
             for( pchIn = asz[ y + 1 ]; *pchIn; pchIn++ )
                 *pch++ = *pchIn;
+#if USE_TIMECONTROL
+	if ((y==3) && asz[7])
+            for( pchIn = asz[ 7 ]; *pchIn; pchIn++ )
+                *pch++ = *pchIn;
+#endif
 
         *pch++ = '\n';
     }
@@ -164,7 +179,7 @@ static char *DrawBoardStd( char *sz, int anBoard[ 2 ][ 25 ], int fRoll,
     *pch++ = '\n';
     
     *pch++ = fRoll ? 'v' : '^';
-    strcpy( pch, "|                  |BAR|                  |     " );
+    strcpy( pch, _("|                  |BAR|                  |     ") );
     pch += 48;
     
     if( asz[ 3 ] )
@@ -241,6 +256,11 @@ static char *DrawBoardStd( char *sz, int anBoard[ 2 ][ 25 ], int fRoll,
             for( pchIn = asz[ 5 - y ]; *pchIn; pchIn++ )
                 *pch++ = *pchIn;
         
+#if USE_TIMECONTROL
+	if ((y==3) && asz[8])
+            for( pchIn = asz[ 8 ]; *pchIn; pchIn++ )
+                *pch++ = *pchIn;
+#endif
         *pch++ = '\n';
     }
 
@@ -261,27 +281,29 @@ static char *DrawBoardStd( char *sz, int anBoard[ 2 ][ 25 ], int fRoll,
 
 /*
  *     GNU Backgammon  Position ID: 0123456789ABCD
- *     +24-23-22-21-20-19------18-17-16-15-14-13-+  O: gnubg (0 points)
- *  OO | O     O  O  O  O |   |                  |  Cube: 2
- *  OO |          O     O |   |                  |  On roll
- *   O |          O       |   |                  |
- *   O |                  |   |                  |
- *   O |                  |   |                  | 
+ *                     Match ID   : 0123456789ABCD
+ *     +24-23-22-21-20-19------18-17-16-15-14-13-+  O: gnubg (Cube: 2)     
+ *  OO | O     O  O  O  O |   |                  |  0 points               
+ *  OO |          O     O |   |                  |  Cube offered at 2      
+ *   O |          O       |   |                  |                         
+ *   O |                  |   |                  |                         
+ *   O |                  |   |                  |                         
  *     |                  |BAR|                  |v Cube: 1 (7 point match)
- *   X |                  |   |                  | 
- *   X |                  |   |                  |
- *   X |                  |   |                  |
- *   X | X  X  X  X       |   |                  |  Rolled 11
- *  XX | X  X  X  X  X    |   |                  |  Cube: 2
- *     +-1--2--3--4--5--6-------7--8--9-10-11-12-+  X: Gary (0 points)
+ *   X |                  |   |                  |                         
+ *   X |                  |   |                  |                         
+ *   X |                  |   |                  |                         
+ *   X | X  X  X  X       |   |                  |  Rolled 11              
+ *  XX | X  X  X  X  X    |   |                  |  0 points               
+ *     +-1--2--3--4--5--6-------7--8--9-10-11-12-+  X: Gary (Cube: 2)      
  *
  */
 
 static char *DrawBoardCls( char *sz, int anBoard[ 2 ][ 25 ], int fRoll,
-                           char *asz[] ) {
+                           char *asz[], char *szMatchID,
+                           int nChequers) {
 
     char *pch = sz, *pchIn;
-    int x, y, an[ 2 ][ 25 ], cOffO = 15, cOffX = 15;
+    int x, y, an[ 2 ][ 25 ], cOffO = nChequers, cOffX = nChequers;
     static char achX[ 16 ] = "     X6789ABCDEF",
         achO[ 16 ] = "     O6789ABCDEF";
 
@@ -290,9 +312,9 @@ static char *DrawBoardCls( char *sz, int anBoard[ 2 ][ 25 ], int fRoll,
         cOffX -= anBoard[ 1 ][ x ];
     }
     
-    strcpy( pch, "    GNU Backgammon  Position ID: " );
+    strcpy( pch, _("    GNU Backgammon  Position ID: ") );
 
-    pch += 30;
+    pch += 33;
 
     if( fRoll )
         strcpy( pch, PositionID( anBoard ) );
@@ -308,6 +330,13 @@ static char *DrawBoardCls( char *sz, int anBoard[ 2 ][ 25 ], int fRoll,
     pch += 14;
     *pch++ = '\n';
             
+    /* match id */
+
+    if ( szMatchID && *szMatchID ) {
+      sprintf ( pch, _("                    Match ID   : %s\n"), szMatchID );
+      pch = strchr ( pch, 0 );
+    }
+            
     strcpy( pch, fRoll ? "    +24-23-22-21-20-19------18-17-16-15-14-13-+  " :
             "    +-1--2--3--4--5--6-------7--8--9-10-11-12-+  " );
     pch += 49;
@@ -389,7 +418,7 @@ static char *DrawBoardCls( char *sz, int anBoard[ 2 ][ 25 ], int fRoll,
 
     *pch++ = '\n';
     
-    strcpy( pch, "    |                  |BAR|                  |" );
+    strcpy( pch, _("    |                  |BAR|                  |") );
     pch += 47;
     *pch++ = fRoll ? 'v' : '^';
     *pch++ = ' ';   
@@ -486,11 +515,13 @@ static char *DrawBoardCls( char *sz, int anBoard[ 2 ][ 25 ], int fRoll,
 }
 
 extern char *DrawBoard( char *sz, int anBoard[ 2 ][ 25 ], int fRoll,
-                        char *asz[] ) {
+                        char *asz[], char *szMatchID, int nChequers ) {
     if( fClockwise == FALSE ) 
-        return ( DrawBoardStd( sz, anBoard, fRoll, asz ) ) ;
+        return ( DrawBoardStd( sz, anBoard, fRoll, asz, 
+                               szMatchID, nChequers  ) ) ;
 
-    return ( DrawBoardCls( sz, anBoard, fRoll, asz ) ) ;   
+    return ( DrawBoardCls( sz, anBoard, fRoll, asz, 
+                           szMatchID, nChequers ) ) ;   
 };
 
 static char *FormatPoint( char *pch, int n ) {
@@ -498,11 +529,11 @@ static char *FormatPoint( char *pch, int n ) {
     assert( n >= 0 );
     
     if( !n ) {
-        strcpy( pch, "off" );
-        return pch + 3;
+        strcpy( pch, _("off") );
+        return pch + strlen(_("off") );
     } else if( n == 25 ) {
-        strcpy( pch, "bar" );
-        return pch + 3;
+        strcpy( pch, _("bar") );
+        return pch + strlen(_("bar") );
     } else if( n > 9 )
         *pch++ = n / 10 + '0';
 
@@ -545,11 +576,11 @@ extern char *FormatMovePlain( char *sz, int anBoard[ 2 ][ 25 ],
         }
         
         if( i < 6 )
-            *pch++ = ' ';
+            *pch++ = ' '; 
     }
 
     *pch = 0;
-    
+
     return sz;
 }
 
@@ -567,18 +598,19 @@ extern void CanonicalMoveOrder( int an[] ) {
 
     int i;
 
-    for( i = 0; i < 8 && an[ i ] > 0; i++ )
+    for( i = 0; i < 4 && an[ 2 * i ] > -1; i++ )
 	;
     
-    qsort( an, i >> 1, sizeof( int ) << 1, CompareMoves );
+    qsort( an, i, sizeof( int ) << 1, CompareMoves );
 }
 
 extern char *FormatMove( char *sz, int anBoard[ 2 ][ 25 ], int anMove[ 8 ] ) {
 
     char *pch = sz;
-    int aanMove[ 4 ][ 4 ], *pnSource[ 4 ], *pnDest[ 4 ], i, j, acRepeat[ 4 ];
+    int aanMove[ 4 ][ 4 ], *pnSource[ 4 ], *pnDest[ 4 ], i, j;
     int fl = 0;
-    
+    int anCount[4], nMoves, nDuplicate, k;
+  
     /* Re-order moves into 2-dimensional array. */
     for( i = 0; i < 4 && anMove[ i << 1 ] >= 0; i++ ) {
         aanMove[ i ][ 0 ] = anMove[ i << 1 ] + 1;
@@ -619,15 +651,56 @@ extern char *FormatMove( char *sz, int anBoard[ 2 ][ 25 ], int anMove[ 8 ] ) {
                 pnSource[ i ] = pnSource[ j ];
                 pnDest[ i ] = pnDest[ j ];
             }
-            
-            acRepeat[ i++ ] = 1;
+
+	    i++;
+        }
+
+    while( i < 4 )
+        pnSource[ i++ ] = NULL;
+
+    for ( i = 0; i < 4; i++)
+        anCount[i] = pnSource[i] ? 1 : 0;
+
+    for ( i = 0; i < 3; i++) {
+        if (pnSource[i]) {
+            nMoves = pnDest[i] - pnSource[i];
+            for (j = i + 1; j < 4; j++) {
+                if (pnSource[j]) {
+                    nDuplicate = 1;
+		    
+                    if (pnDest[j] - pnSource[j] != nMoves)
+                        nDuplicate = 0;
+                    else
+                        for (k = 0; k <= nMoves && nDuplicate; k++)
+			    {
+				if (pnSource[i][k] != pnSource[j][k])
+				    nDuplicate = 0;
+			    }
+                    if (nDuplicate) {
+                        anCount[i]++;
+                        pnSource[j] = NULL;
+                    }
+                }
+            }
+        }
+    }
+
+    /* Compact array. */
+    i = 0;
+
+    for( j = 0; j < 4; j++ )
+        if( pnSource[ j ] ) {
+            if( j > i ) {
+                pnSource[ i ] = pnSource[ j ];
+                pnDest[ i ] = pnDest[ j ];
+		anCount[ i ] = anCount[ j ];
+            }
+
+	    i++;
         }
 
     if( i < 4 )
         pnSource[ i ] = NULL;
-    
-    /* FIXME Combine repeated moves, e.g. 6/5 6/5 -> 6/5(2).  Then compact
-       again. */
 
     for( i = 0; i < 4 && pnSource[ i ]; i++ ) {
         if( i )
@@ -639,6 +712,7 @@ extern char *FormatMove( char *sz, int anBoard[ 2 ][ 25 ], int anMove[ 8 ] ) {
             *pch = '/';
             pch = FormatPoint( pch + 1, pnSource[ i ][ j ] );
             *pch++ = '*';
+            fl |= 1 << pnSource[ i ][ j ];
         }
 
         *pch = '/';
@@ -649,6 +723,12 @@ extern char *FormatMove( char *sz, int anBoard[ 2 ][ 25 ], int anMove[ 8 ] ) {
             *pch++ = '*';
             fl |= 1 << *pnDest[ i ];
         }
+	
+        if (anCount[i] > 1) {
+            *pch++ = '(';
+            *pch++ = '0' + anCount[i];
+            *pch++ = ')';
+        }
     }
 
     *pch = 0;
@@ -834,24 +914,42 @@ extern int ParseMove( char *pch, int an[ 8 ] ) {
     }
 
     if( i < 8 )
-	an[ i ] = 0;
+      an[ i ] = -1;
 
     CanonicalMoveOrder( an );
     
     return i >> 1;
 }
 
+/*
+ * Generate FIBS board
+ *
+ * (see <URL:http://www.fibs.com/fibs_interface.html#board_state>)
+ *
+ * gnubg extends the meaning of the "doubled" field, so the sign of
+ * the field indicates who doubled. This is necessaty to correctly
+ * detects beavers and raccoons.
+ *
+ */
+
 extern char *FIBSBoard( char *pch, int anBoard[ 2 ][ 25 ], int fRoll,
 			char *szPlayer, char *szOpp, int nMatchTo,
 			int nScore, int nOpponent, int nDice0, int nDice1,
 			int nCube, int fCubeOwner, int fDoubled, int fTurn,
-			int fCrawford ) {
+			int fCrawford, int nChequers ) {
     char *sz = pch;
     int i, anOff[ 2 ];
     
     /* Names and match length/score */
-    sprintf( sz, "board:%s:%s:%d:%d:%d:", szPlayer, szOpp, nMatchTo, nScore,
-	     nOpponent );
+    strcpy( sz, "board:" );
+    
+    for( sz += 6; *szPlayer; szPlayer++ )
+	*sz++ = ( *szPlayer != ':' ? *szPlayer : '_' );
+    
+    for( *sz++ = ':'; *szOpp; szOpp++ )    
+	*sz++ = ( *szOpp != ':' ? *szOpp : '_' );
+    
+    sprintf( sz, ":%d:%d:%d:", nMatchTo, nScore, nOpponent );
 
     /* Opponent on bar */
     sprintf( strchr( sz, 0 ), "%d:", -anBoard[ 0 ][ 24 ] );
@@ -867,7 +965,7 @@ extern char *FIBSBoard( char *pch, int anBoard[ 2 ][ 25 ], int fRoll,
     /* Whose turn */
     strcat( strchr( sz, 0 ), fRoll ? "1:" : "-1:" );
 
-    anOff[ 0 ] = anOff[ 1 ] = 15;
+    anOff[ 0 ] = anOff[ 1 ] = nChequers ? nChequers : 15;
     for( i = 0; i < 25; i++ ) {
 	anOff[ 0 ] -= anBoard[ 0 ][ i ];
 	anOff[ 1 ] -= anBoard[ 1 ][ i ];
@@ -947,7 +1045,5 @@ extern int ParseFIBSBoard( char *pch, int anBoard[ 2 ][ 25 ],
 
     *pfCubeOwner = fCanDouble != fOppCanDouble ? fCanDouble : -1;
 
-    *pfDoubled = *pfDoubled != 0;
-
     return 0;
 }
