@@ -1,7 +1,22 @@
 /*
  * neuralnet.c
  *
- * by Gary Wong, 1998-2000
+ * by Gary Wong <gtw@gnu.org>, 1998, 1999, 2000, 2001, 2002.
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of version 2 of the GNU General Public License as
+ * published by the Free Software Foundation.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *
+ * $Id: neuralnet.c,v 1.19 2003/12/29 20:29:33 uid65656 Exp $
  */
 
 #include "config.h"
@@ -16,6 +31,25 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <time.h>
+#include <assert.h>
+
+#define SIGMOID_BAUR 0
+#define SIGMOID_JTH 0
+
+#if HAVE_LIBATLAS
+#warning "LIBATLAS processing..."
+#include <cblas.h>
+#endif /* HAVE_LIBATLAS */
+
+#if SIGMOID_BAUR
+#define sigmoid sigmoid_baur
+#elif SIGMOID_JTH
+#define sigmoid sigmoid_jth
+#else
+#define sigmoid sigmoid_original
+#endif
+
 
 /* e[k] = exp(k/10) / 10 */
 static float e[100] = {
@@ -132,7 +166,8 @@ static float e[100] = {
    The Intel x87's `f2xm1' instruction makes calculating accurate
    exponentials comparatively fast, but still about 30% slower than
    the lookup table used here. */
-static inline float sigmoid(float const xin) {
+//static inline float sigmoid(float const xin) {
+static float sigmoid_original(float const xin) {
     
     if( !signbit( xin ) ) { /* signbit() can be faster than a compare to 0.0 */
 	/* xin is almost always positive; we place this branch of the `if'
@@ -157,6 +192,119 @@ static inline float sigmoid(float const xin) {
     }
 }
 
+#if SIGMOID_BAUR
+
+#define SIG_Q 10.0f     /* reciprocal of precision step */                      
+#define SIG_MAX 100     /* half number of entries in lookup table */
+/* note: the lookup table covers the interval [ -SIG_MAX/SIG_Q  to
++SIG_MAX/SIG_Q ] */
+static float Sig[2*SIG_MAX+1];
+
+#endif /* SIGMOID_BAUR */
+
+#if SIGMOID_JTH
+
+#define SIG_Q 10.0f
+#define SIG_MAX 100  /* half number of entries */
+
+static float SigD0[ 2 * SIG_MAX + 1 ]; /* sigmoid(x0) */
+static float SigD1[ 2 * SIG_MAX + 1 ]; /* sigmoid'(x0) */
+
+#endif /* SIGMOID_JTH */
+
+extern void 
+ComputeSigTable (void) {
+
+#if SIGMOID_BAUR
+
+  int i;
+
+  for (i = -SIG_MAX; i < SIG_MAX+1; i++) {
+    float x = (float) i / SIG_Q;
+    /* Sig[SIG_MAX+i] = 1.0f / (1.0f + exp (x));  more accurate, 
+                                                  but fails gnubgtest */
+    /* use the current sigmoid function instead :-)  */
+    Sig[SIG_MAX+i] = sigmoid_original(x);  
+  }
+
+  printf( "Table with %d elements initialised\n", 2 * SIG_MAX + 1 );
+
+#endif /* SIGMOID_BAUR */
+
+#if SIGMOID_JTH
+
+  int i;
+  for ( i = -SIG_MAX; i < SIG_MAX + 1; ++i ) {
+    float x = (float) i / SIG_Q;
+    SigD0[ SIG_MAX + i ] = sigmoid_original(x);
+    SigD1[ SIG_MAX + i ] = 
+      - exp(x) * SigD0[ SIG_MAX + i ] * SigD0[ SIG_MAX + i ] / SIG_Q;
+  }
+
+  printf( "SIGMOID_JTH: table with %d elements initialised.\n", 
+          2 * SIG_MAX + 1 );
+
+#endif /* SIGMOID_JTH */
+
+}
+
+#if SIGMOID_BAUR
+
+static float 
+sigmoid_baur (float const x) {
+
+  float x2 = x * SIG_Q;
+  int i = x2;
+
+  if (i > - SIG_MAX) {
+    if (i < SIG_MAX) {
+      float a = Sig[i+SIG_MAX];
+      float b = Sig[i+SIG_MAX+1];
+      return a + (b - a) * (x2 - i);
+    }
+    else
+      /* warning: this is 1.0f/(1.0f+exp(9.9f)) */
+      return 1.0f / 19931.370438230298f;  
+  }
+  else
+    /* warning: this is 1.0f/(1.0f+exp(-9.9f)) */
+    return 19930.370438230298f / 19931.370438230298f;  
+}
+
+#endif /* SIGMOID_BAUR */
+
+#if SIGMOID_JTH
+
+static float
+sigmoid_jth( const float x ) {
+
+  float x2 = x * SIG_Q;
+  int i = x2;
+
+  if (i > - SIG_MAX) {
+    if (i < SIG_MAX) {
+      float fx0   = SigD0[i+SIG_MAX];
+      float dfx0 = SigD1[i+SIG_MAX];
+      float s = fx0 + dfx0 * ( x2 - i );
+      /*
+      float d = sigmoid_original(x);
+      if ( fabs(s-d) > 2.0e-3 ) {
+        printf( "%f %f %f %f %f %f %f \n", x, x2, x0, fx0, dfx0, s, d );
+        assert(0);
+        }*/
+      return s;
+    }
+    else
+      /* warning: this is 1.0f/(1.0f+exp(9.9f)) */
+      return 1.0f / 19931.370438230298f;  
+  }
+  else
+    /* warning: this is 1.0f/(1.0f+exp(-9.9f)) */
+    return 19930.370438230298f / 19931.370438230298f;  
+}
+
+#endif /* SIGMOID_JTH */
+
 
 /*  #define sigmoid( x ) ( (x) > 0.0f ? \ */
 /*  		       1.0f / ( 2.0f + (x) + (1.0f / 2.0f) * (x) * (x) ) : \ */
@@ -169,8 +317,14 @@ static randctx rc; /* for irand */
 static void CheckRC( void ) {
 
     if( !frc ) {
-	irandinit( &rc, FALSE );
-	frc = TRUE;
+        int i;
+
+        rc.randrsl[ 0 ] = time( NULL );
+        for( i = 0; i < RANDSIZ; i++ )
+           rc.randrsl[ i ] = rc.randrsl[ 0 ];
+        irandinit( &rc, TRUE );
+
+        frc = TRUE;
     }
 }
 
@@ -186,7 +340,8 @@ extern int NeuralNetCreate( neuralnet *pnn, int cInput, int cHidden,
     pnn->rBetaHidden = rBetaHidden;
     pnn->rBetaOutput = rBetaOutput;
     pnn->nTrained = 0;
-
+    pnn->fDirect = FALSE;
+   
     if( !( pnn->arHiddenWeight = malloc( cHidden * cInput *
 					 sizeof( float ) ) ) )
 	return -1;
@@ -211,7 +366,10 @@ extern int NeuralNetCreate( neuralnet *pnn, int cInput, int cHidden,
     }
 
     CheckRC();
-    
+
+    pnn->savedBase = malloc( cHidden * sizeof( float ) ); 
+    pnn->savedIBase = malloc( cInput * sizeof( float ) ); 
+ 
     for( i = cHidden * cInput, pf = pnn->arHiddenWeight; i; i-- )
 	*pf++ = ( (int) ( irand( &rc ) & 0xFFFF ) - 0x8000 ) / 131072.0;
     
@@ -226,19 +384,58 @@ extern int NeuralNetCreate( neuralnet *pnn, int cInput, int cHidden,
 
     return 0;
 }
+extern void *NeuralNetCreateDirect( neuralnet *pnn, void *p ) {
+ 
+   pnn->cInput = *( ( (int *) p )++ );
+   pnn->cHidden = *( ( (int *) p )++ );
+   pnn->cOutput = *( ( (int *) p )++ );
+   pnn->nTrained = *( ( (int *) p )++ );
+   pnn->fDirect = TRUE;
+   pnn->rBetaHidden = *( ( (float *) p )++ );
+   pnn->rBetaOutput = *( ( (float *) p )++ );
+ 
+   if( pnn->cInput < 1 || pnn->cHidden < 1 || pnn->cOutput < 1 ||
+     pnn->nTrained < 0 || pnn->rBetaHidden <= 0.0 ||
+     pnn->rBetaOutput <= 0.0 ) {
+     errno = EINVAL;
+     
+     return NULL;
+   }
+ 
+   pnn->arHiddenWeight = p;
+   ( (float *) p ) += pnn->cInput * pnn->cHidden;
+   pnn->arOutputWeight = p;
+   ( (float *) p ) += pnn->cHidden * pnn->cOutput;
+   pnn->arHiddenThreshold = p;
+   ( (float *) p ) += pnn->cHidden;
+   pnn->arOutputThreshold = p;
+   ( (float *) p ) += pnn->cOutput;
 
-extern int NeuralNetDestroy( neuralnet *pnn ) {
+   pnn->savedBase = malloc( pnn->cHidden * sizeof( float ) ); 
+   pnn->savedIBase = malloc( pnn->cInput * sizeof( float ) ); 
 
-    free( pnn->arHiddenWeight );
-    free( pnn->arOutputWeight );
-    free( pnn->arHiddenThreshold );
-    free( pnn->arOutputThreshold );
-
-    return 0;
+   return p;
 }
 
-static int Evaluate( neuralnet *pnn, float arInput[], float ar[],
-		     float arOutput[] ) {
+extern int
+NeuralNetDestroy( neuralnet *pnn )
+{
+  if( !pnn->fDirect ) {
+    free( pnn->arHiddenWeight ); pnn->arHiddenWeight = 0;
+    free( pnn->arOutputWeight ); pnn->arOutputWeight = 0;
+    free( pnn->arHiddenThreshold ); pnn->arHiddenThreshold = 0;
+    free( pnn->arOutputThreshold ); pnn->arOutputThreshold = 0;
+  }
+
+  free(pnn->savedBase); pnn->savedBase = 0;
+  free(pnn->savedIBase); pnn->savedIBase = 0;
+  
+  return 0;
+}
+
+static int EvaluateOld( neuralnet *pnn, float arInput[], float ar[],
+                        float arOutput[], float *saveAr ) {
+
     int i, j;
     float *prWeight;
 
@@ -263,6 +460,114 @@ static int Evaluate( neuralnet *pnn, float arInput[], float ar[],
 	} else
 	    prWeight += pnn->cHidden;
     }
+
+    if( saveAr)
+      memcpy( saveAr, ar, pnn->cHidden * sizeof( *saveAr));
+
+
+    for( i = 0; i < pnn->cHidden; i++ )
+	ar[ i ] = sigmoid( -pnn->rBetaHidden * ar[ i ] );
+
+    /* Calculate activity at output nodes */
+    prWeight = pnn->arOutputWeight;
+
+    for( i = 0; i < pnn->cOutput; i++ ) {
+	float r = pnn->arOutputThreshold[ i ];
+	
+	for( j = 0; j < pnn->cHidden; j++ )
+	    r += ar[ j ] * *prWeight++;
+
+	arOutput[ i ] = sigmoid( -pnn->rBetaOutput * r );
+    }
+
+    return 0;
+
+}
+
+static int Evaluate( neuralnet *pnn, float arInput[], float ar[],
+		     float arOutput[], float *saveAr ) {
+  int i;
+
+#if !HAVE_LIBATLAS
+  return EvaluateOld( pnn, arInput, ar, arOutput, saveAr );
+#else /* HAVE_LIBATLAS */
+
+    /* BLAS implementation */
+
+    /* ar = t(hidden) */
+    memcpy( ar, pnn->arHiddenThreshold, 
+            pnn->cHidden * sizeof ( float ) );
+
+    /* activity at hidden nodes: ar = W(hidden) * i + t(hidden) */
+
+    cblas_sgemv( CblasColMajor, CblasNoTrans,
+                 pnn->cHidden, pnn->cInput, 1.0f,
+                 pnn->arHiddenWeight, pnn->cHidden,
+                 arInput, 1,
+                 1.0f, ar, 1 );
+
+    /* save result for later use */
+    
+    if( saveAr)
+      memcpy( saveAr, ar, pnn->cHidden * sizeof( *saveAr));
+
+    /* apply sigmoid */
+
+    for( i = 0; i < pnn->cHidden; i++ )
+	ar[ i ] = sigmoid( -pnn->rBetaHidden * ar[ i ] );
+
+    /* calculate output: arOutput = W(output) * ar + t(output) */
+
+    memcpy( arOutput, pnn->arOutputThreshold, 
+            pnn->cOutput * sizeof ( float ) );
+
+    cblas_sgemv( CblasRowMajor, CblasNoTrans,
+                 pnn->cOutput, pnn->cHidden, 1.0f,
+                 pnn->arOutputWeight, pnn->cHidden,
+                 ar, 1,
+                 1.0f, arOutput, 1 );
+
+    /* apply sigmoid */
+
+    for( i = 0; i < pnn->cOutput; i++ )
+	arOutput[ i ] = sigmoid( -pnn->rBetaOutput * arOutput[ i ] );
+
+#endif /* HAVE_LIBATLAS */
+
+    return 0;
+}
+
+static int EvaluateFromBaseOld( neuralnet *pnn, float arInputDif[], float ar[],
+		     float arOutput[] ) {
+
+    int i, j;
+    float *prWeight;
+
+    /* Calculate activity at hidden nodes */
+/*    for( i = 0; i < pnn->cHidden; i++ )
+	ar[ i ] = pnn->arHiddenThreshold[ i ]; */
+
+    prWeight = pnn->arHiddenWeight;
+    
+    for( i = 0; i < pnn->cInput; ++i ) {
+	float const ari = arInputDif[ i ];
+
+	if( ari ) {
+	    float *pr = ar;
+
+	    if( ari == 1.0f )
+		for( j = pnn->cHidden; j; j-- )
+		    *pr++ += *prWeight++;
+	    else
+            if(ari == -1.0f)
+              for(j = pnn->cHidden; j; j-- ) 
+                *pr++ -= *prWeight++;
+            else
+		for( j = pnn->cHidden; j; j-- )
+		    *pr++ += *prWeight++ * ari;
+	} else
+	    prWeight += pnn->cHidden;
+    }
     
     for( i = 0; i < pnn->cHidden; i++ )
 	ar[ i ] = sigmoid( -pnn->rBetaHidden * ar[ i ] );
@@ -279,11 +584,57 @@ static int Evaluate( neuralnet *pnn, float arInput[], float ar[],
 	arOutput[ i ] = sigmoid( -pnn->rBetaOutput * r );
     }
 
+    return 0;
+
+}
+
+
+static int EvaluateFromBase( neuralnet *pnn, float arInputDif[], float ar[],
+		     float arOutput[] ) {
+    int i;
+
+#if !HAVE_LIBATLAS
+    return EvaluateFromBaseOld( pnn, arInputDif, ar, arOutput );
+#else /* HAVE_LIBATLAS */
+
+    /* BLAS implementation */
+
+    /* activity at hidden nodes: ar = W(hidden) * i + t(hidden) */
+
+    cblas_sgemv( CblasColMajor, CblasNoTrans,
+                 pnn->cHidden, pnn->cInput, 1.0f,
+                 pnn->arHiddenWeight, pnn->cHidden,
+                 arInputDif, 1,
+                 1.0f, ar, 1 );
+    
+    /* apply sigmoid */
+
+    for( i = 0; i < pnn->cHidden; i++ )
+	ar[ i ] = sigmoid( -pnn->rBetaHidden * ar[ i ] );
+
+    /* calculate output: arOutput = W(output) * ar + t(output) */
+
+    memcpy( arOutput, pnn->arOutputThreshold, 
+            pnn->cOutput * sizeof ( float ) );
+
+    cblas_sgemv( CblasRowMajor, CblasNoTrans,
+                 pnn->cOutput, pnn->cHidden, 1.0f,
+                 pnn->arOutputWeight, pnn->cHidden,
+                 ar, 1,
+                 1.0f, arOutput, 1 );
+
+    /* apply sigmoid */
+
+    for( i = 0; i < pnn->cOutput; i++ )
+	arOutput[ i ] = sigmoid( -pnn->rBetaOutput * arOutput[ i ] );
+    
+#endif /* HAVE_LIBATLAS */
+
     return 0;
 }
 
 extern int NeuralNetEvaluate( neuralnet *pnn, float arInput[],
-			      float arOutput[] ) {
+			      float arOutput[], NNEvalType t ) {
 #if __GNUC__
     float ar[ pnn->cHidden ];
 #elif HAVE_ALLOCA
@@ -291,8 +642,41 @@ extern int NeuralNetEvaluate( neuralnet *pnn, float arInput[],
 #else
     float ar[ 1024 ];
 #endif
-    
-    return Evaluate( pnn, arInput, ar, arOutput );
+    switch( t ) {
+      case NNEVAL_NONE:
+      {
+        Evaluate(pnn, arInput, ar, arOutput, 0);
+        break;
+      }
+      case NNEVAL_SAVE:
+      {
+        memcpy(pnn->savedIBase, arInput, pnn->cInput * sizeof(*ar));
+        Evaluate(pnn, arInput, ar, arOutput, pnn->savedBase);
+        break;
+      }
+      case NNEVAL_FROMBASE:
+      {
+        int i;
+        
+        memcpy(ar, pnn->savedBase, pnn->cHidden * sizeof(*ar));
+  
+        {
+          float* r = arInput;
+          float* s = pnn->savedIBase;
+         
+          for(i = 0; i < pnn->cInput; ++i, ++r, ++s) {
+            if( *r != *s ) {
+              *r -= *s;
+            } else {
+              *r = 0.0;
+            }
+          }
+        }
+        EvaluateFromBase(pnn, arInput, ar, arOutput);
+        break;
+      }
+    }
+    return 0;
 }
 
 /*
@@ -316,7 +700,7 @@ extern int NeuralNetDifferentiate( neuralnet *pnn, float arInput[],
 #endif
     int i, j;
 
-    Evaluate( pnn, arInput, ar, arOutput );
+    Evaluate( pnn, arInput, ar, arOutput, 0 );
 
     memcpy( arIDelta, arInput, sizeof( arIDelta ) );
     
@@ -325,7 +709,7 @@ extern int NeuralNetDifferentiate( neuralnet *pnn, float arInput[],
 	if( i )
 	    arIDelta[ i - 1 ] = arInput[ i - 1 ];
 	
-	Evaluate( pnn, arIDelta, ar, arODelta );
+	Evaluate( pnn, arIDelta, ar, arODelta, 0 );
 
 	for( j = 0; j < pnn->cOutput; j++ )
 	    arDerivative[ j * pnn->cInput + i ] =
@@ -416,7 +800,7 @@ extern int NeuralNetTrain( neuralnet *pnn, float arInput[], float arOutput[],
     float ar[ 1024 ], arOutputError[ 128 ], arHiddenError[ 1024 ];
 #endif
     
-    Evaluate( pnn, arInput, ar, arOutput );
+    Evaluate( pnn, arInput, ar, arOutput, 0 );
 
     /* Calculate error at output nodes */
     for( i = 0; i < pnn->cOutput; i++ )
@@ -580,12 +964,18 @@ extern int NeuralNetLoad( neuralnet *pnn, FILE *pf ) {
 	if( fscanf( pf, "%f\n", pr++ ) < 1 )
 	    return -1;
 
+    for(i = 0; i < pnn->cHidden; ++i)
+      pnn->savedBase[i] = 0.0;
+
+    for(i = 0; i < pnn->cInput; ++i) 
+      pnn->savedIBase[i] = 0.0;
+ 
     return 0;
 }
 
 extern int NeuralNetLoadBinary( neuralnet *pnn, FILE *pf ) {
 
-    int nTrained;
+    int nTrained, i;
 
 #define FREAD( p, c ) \
     if( fread( (p), sizeof( *(p) ), (c), pf ) < (c) ) return -1;
@@ -616,6 +1006,12 @@ extern int NeuralNetLoadBinary( neuralnet *pnn, FILE *pf ) {
     FREAD( pnn->arOutputThreshold, pnn->cOutput );
 #undef FREAD
 
+    for(i = 0; i < pnn->cHidden; ++i)
+      pnn->savedBase[i] = 0.0;
+
+    for(i = 0; i < pnn->cInput; ++i) 
+      pnn->savedIBase[i] = 0.0;
+ 
     return 0;
 }
 
