@@ -32,19 +32,17 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
- * $Id: positionid.c,v 1.9 2001/04/24 12:35:50 thyssen Exp $
+ * $Id: positionid.c,v 1.29 2003/08/18 22:21:30 joseph Exp $
  */
 
 #include <assert.h>
+#include <stdio.h>
 #include <errno.h>
+#include <string.h>
 #include "positionid.h"
 
-static
-#if defined( __GNUC__ )
-inline
-#endif
-void
-addBits(unsigned char auchKey[10], int bitPos, int nBits)
+static inline void
+addBits(unsigned char auchKey[10], int const bitPos, int const nBits)
 {
   int const k = bitPos / 8;
   int const r = (bitPos & 0x7);
@@ -84,9 +82,9 @@ PositionKey(int anBoard[2][25], unsigned char auchKey[10])
   }
 }
 
-extern char *PositionIDFromKey( unsigned char auchKey[ 10 ] ) {
+extern char *PositionIDFromKey( const unsigned char auchKey[ 10 ] ) {
 
-    unsigned char *puch = auchKey;
+    const unsigned char *puch = auchKey;
     static char szID[ 15 ];
     char *pch = szID;
     static char aszBase64[ 64 ] =
@@ -121,43 +119,85 @@ extern char *PositionID( int anBoard[ 2 ][ 25 ] ) {
     return PositionIDFromKey( auch );
 }
 
-
-extern int CheckPosition( int anBoard[ 2 ][ 25 ] ) {
-
+extern int
+CheckPosition( int anBoard[ 2 ][ 25 ] )
+{
     int ac[ 2 ], i;
 
+    /* Check for a point with a negative number of chequers */
+    for( i = 0; i < 25; i++ )
+	if( anBoard[ 0 ][ i ] < 0 ||
+	    anBoard[ 1 ][ i ] < 0 ) {
+            errno = EINVAL;
+            return 0;
+	}
+    
     /* Check for a player with over 15 chequers */
     for( i = ac[ 0 ] = ac[ 1 ] = 0; i < 25; i++ )
         if( ( ac[ 0 ] += anBoard[ 0 ][ i ] ) > 15 ||
             ( ac[ 1 ] += anBoard[ 1 ][ i ] ) > 15 ) {
             errno = EINVAL;
-            return -1;
+            return 0;
         }
 
     /* Check for both players having chequers on the same point */
     for( i = 0; i < 24; i++ )
         if( anBoard[ 0 ][ i ] && anBoard[ 1 ][ 23 - i ] ) {
             errno = EINVAL;
-            return -1;
+            return 0;
         }
 
     /* Check for both players on the bar against closed boards */
     for( i = 0; i < 6; i++ )
         if( anBoard[ 0 ][ i ] < 2 || anBoard[ 1 ][ i ] < 2 )
-            return 0;
+            return 1;
 
     if( !anBoard[ 0 ][ 24 ] || !anBoard[ 1 ][ 24 ] )
-        return 0;
+        return 1;
     
     errno = EINVAL;
-    return -1;
+    return 0;
+}
+
+extern void ClosestLegalPosition( int anBoard[ 2 ][ 25 ] ) {
+
+    int i, j, ac[ 2 ] = { 15, 15 };
+
+    /* Force a non-negative number of chequers on all points */
+    for( i = 0; i < 2; i++ )
+	for( j = 0; j < 25; j++ )
+	    if( anBoard[ i ][ j ] < 0 )
+		anBoard[ i ][ j ] = 0;
+
+    /* Limit each player to 15 chequers */
+    for( i = 0; i < 2; i++ )
+	for( j = 0; j < 25; j++ )
+	    if( ( ac[ i ] -= anBoard[ i ][ j ] ) < 0 ) {
+		anBoard[ i ][ j ] += ac[ i ];
+		ac[ i ] = 0;
+	    }
+
+    /* Forbid both players having a chequer on the same point */
+    for( i = 0; i < 24; i++ )
+	if( anBoard[ 0 ][ i ] )
+	    anBoard[ 1 ][ 23 - i ] = 0;
+
+    /* If both players have closed boards, let at least one of them off
+       the bar */
+    for( i = 0; i < 6; i++ )
+        if( anBoard[ 0 ][ i ] < 2 || anBoard[ 1 ][ i ] < 2 )
+	    /* open board */
+            return;
+
+    if( anBoard[ 0 ][ 24 ] )
+	anBoard[ 1 ][ 24 ] = 0;
 }
 
 extern void
-PositionFromKey(int anBoard[2][25], unsigned char* pauch)
+PositionFromKey(int anBoard[2][25], const unsigned char* pauch)
 {
   int i = 0, j  = 0, k;
-  unsigned char* a;
+  const unsigned char* a;
 
   memset(anBoard[0], 0, sizeof(anBoard[0]));
   memset(anBoard[1], 0, sizeof(anBoard[1]));
@@ -179,7 +219,8 @@ PositionFromKey(int anBoard[2][25], unsigned char* pauch)
   }
 }
 
-static int Base64( char ch ) {
+extern int
+Base64( const char ch ) {
 
     if( ch >= 'A' && ch <= 'Z' )
         return ch - 'A';
@@ -196,16 +237,17 @@ static int Base64( char ch ) {
     return 63;
 }
 
-extern int PositionFromID( int anBoard[ 2 ][ 25 ], char *pchEnc ) {
-
+extern int
+PositionFromID(int anBoard[ 2 ][ 25 ], const char* pchEnc)
+{
   unsigned char auchKey[ 10 ], ach[ 15 ], *pch = ach, *puch = auchKey;
   int i;
 
+  memset ( ach, 0, 15 );
+
   for( i = 0; i < 14 && pchEnc[ i ]; i++ )
     pch[ i ] = Base64( pchEnc[ i ] );
 
-  pch[ i ] = 0;
-    
   for( i = 0; i < 3; i++ ) {
     *puch++ = ( pch[ 0 ] << 2 ) | ( pch[ 1 ] >> 4 );
     *puch++ = ( pch[ 1 ] << 4 ) | ( pch[ 2 ] >> 2 );
@@ -221,7 +263,8 @@ extern int PositionFromID( int anBoard[ 2 ][ 25 ], char *pchEnc ) {
   return CheckPosition( anBoard );
 }
 
-extern int EqualKeys( unsigned char auch0[ 10 ], unsigned char auch1[ 10 ] ) {
+extern int 
+EqualKeys( const unsigned char auch0[ 10 ], const unsigned char auch1[ 10 ] ) {
 
     int i;
 
@@ -231,7 +274,7 @@ extern int EqualKeys( unsigned char auch0[ 10 ], unsigned char auch1[ 10 ] ) {
     
     return 1;
 }
-
+ 
 extern int EqualBoards( int anBoard0[ 2 ][ 25 ], int anBoard1[ 2 ][ 25 ] ) {
 
     int i;
@@ -244,20 +287,23 @@ extern int EqualBoards( int anBoard0[ 2 ][ 25 ], int anBoard1[ 2 ][ 25 ] ) {
     return 1;
 }
 
-static int anCombination[ 21 ][ 6 ], fCalculated = 0;
+#define MAX_N 40
+#define MAX_R 25
+
+static int anCombination[ MAX_N ][ MAX_R ], fCalculated = 0;
 
 static int InitCombination( void ) {
 
     int i, j;
 
-    for( i = 0; i < 21; i++ )
+    for( i = 0; i < MAX_N; i++ )
         anCombination[ i ][ 0 ] = i + 1;
     
-    for( j = 1; j < 6; j++ )
+    for( j = 1; j < MAX_R; j++ )
         anCombination[ 0 ][ j ] = 0;
 
-    for( i = 1; i < 21; i++ )
-        for( j = 1; j < 6; j++ )
+    for( i = 1; i < MAX_N; i++ )
+        for( j = 1; j < MAX_R; j++ )
             anCombination[ i ][ j ] = anCombination[ i - 1 ][ j - 1 ] +
                 anCombination[ i - 1 ][ j ];
 
@@ -266,20 +312,20 @@ static int InitCombination( void ) {
     return 0;
 }
 
-static int Combination( int n, int r ) {
+extern int Combination( const int n, const int r ) {
 
     assert( n > 0 );
     assert( r > 0 );
-    assert( n <= 21 );
-    assert( r <= 6 );
+    assert( n <= MAX_N );
+    assert( r <= MAX_R );
 
     if( !fCalculated )
         InitCombination();
-    
+
     return anCombination[ n - 1 ][ r - 1 ];
 }
 
-static int PositionF( int fBits, int n, int r ) {
+static int PositionF( const int fBits, const int n, const int r ) {
 
     if( n == r )
         return 0;
@@ -288,21 +334,25 @@ static int PositionF( int fBits, int n, int r ) {
         PositionF( fBits, n - 1, r - 1 ) : PositionF( fBits, n - 1, r );
 }
 
-extern unsigned short PositionBearoff( int anBoard[ 6 ] ) {
+extern 
+unsigned int PositionBearoff( const int anBoard[],
+                              const int nPoints,
+                              const int nChequers ) {
 
     int i, fBits, j;
 
-    for( j = 5, i = 0; i < 6; i++ )
+    for( j = nPoints - 1, i = 0; i < nPoints; i++ )
         j += anBoard[ i ];
 
     fBits = 1 << j;
     
-    for( i = 0; i < 6; i++ ) {
+    for( i = 0; i < nPoints; i++ ) {
         j -= anBoard[ i ] + 1;
         fBits |= ( 1 << j );
+
     }
 
-    return PositionF( fBits, 21, 6 );
+    return PositionF( fBits, nChequers + nPoints, nPoints );
 }
 
 static int PositionInv( int nID, int n, int r ) {
@@ -320,18 +370,40 @@ static int PositionInv( int nID, int n, int r ) {
         PositionInv( nID - nC, n - 1, r - 1 ) : PositionInv( nID, n - 1, r );
 }
 
-extern void PositionFromBearoff( int anBoard[ 6 ], unsigned short usID ) {
+extern void PositionFromBearoff( int anBoard[], const unsigned int usID,
+                                 const int nPoints, const int nChequers ) {
     
-    int fBits = PositionInv( usID, 21, 6 );
+    int fBits = PositionInv( usID, nChequers + nPoints, nPoints );
     int i, j;
 
-    for( i = 0; i < 6; i++ )
+    for( i = 0; i < nPoints; i++ )
         anBoard[ i ] = 0;
     
-    for( j = 5, i = 0; j >= 0 && i < 21; i++ ) {
+    for( j = nPoints - 1, i = 0; j >= 0 && i < ( nChequers + nPoints ); i++ ) {
         if( fBits & ( 1 << i ) )
             j--;
         else
             anBoard[ j ]++;
     }
 }
+
+extern unsigned short
+PositionIndex(int g, int anBoard[6])
+{
+  int i, fBits;
+  int j = g - 1;
+
+  for(i = 0; i < g; i++ )
+    j += anBoard[ i ];
+
+  fBits = 1 << j;
+    
+  for(i = 0; i < g; i++) {
+    j -= anBoard[ i ] + 1;
+    fBits |= ( 1 << j );
+  }
+
+  /* FIXME: 15 should be replaced by nChequers, but the function is
+     only called from bearoffgammon, so this should be fine. */
+  return PositionF( fBits, 15, g );
+}
